﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Xml.Linq;
using SpotnetWatchdog.Entities;

namespace SpotnetWatchdog.Utils
{
    class WatchdogRepository : IWatchdogRepository
    {
        #region Fields

        private List<Category> _categories;
        private string _categoriesFile;
        private object _padlock = new object();
        private string _watchdogFile;
        private string _watchdogFileDirectory;
        private ObservableCollection<WatchItem> _watchItems;

        private const string CATEGORY_ELEMENT = "category";
        private const string DESCRIPTION_ELEMENT = "description";
        private const string SEARCH_STRING_ELEMENT = "searchString";
        private const string SEARCH_FOR_ALL_WORDS_ELEMENT = "searchForAllWords";
        private const string SUB_CATEGORY_ELEMENT = "subCategory";
        private const string SPOTTER_ELEMENT = "spotter";
        private const string WHOLE_WORDS_ONLY_ELEMENT = "wholeWordsOnly";

        #endregion

        #region Constructors

        private WatchdogRepository()
        {
            _categoriesFile = Path.Combine(Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location)
                , "spotnet_categories.xml");

            _watchdogFileDirectory = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "SpotnetWatchdog");
            _watchdogFile = Path.Combine(_watchdogFileDirectory, "spotnet_watchdog.xml");
        }

        #endregion

        #region Public Properties

        public List<Category> Categories
        {
            get
            {
                if (_categories == null)
                {
                    lock (_padlock)
                    {
                        if (_categories == null)
                        {
                            _categories = GetCategories();
                        }
                    }
                }

                return _categories;
            }
        }

        public ObservableCollection<WatchItem> WatchItems
        {
            get
            {
                if (_watchItems == null)
                {
                    lock (_padlock)
                    {
                        if (_watchItems == null)
                        {
                            _watchItems = GetWatchItems();
                        }
                    }
                }

                return _watchItems;
            }
        }

        #endregion

        #region Public Methods

        public void SaveWatchItems()
        {
            if (_watchItems == null) throw new InvalidOperationException();

            if (!Directory.Exists(_watchdogFileDirectory))
            {
                Directory.CreateDirectory(_watchdogFileDirectory);
            }

            XDocument repository = new XDocument(new XElement("spotnetWatchdog"));
                
            foreach (WatchItem watchItem in _watchItems)
            {
                repository.Root.Add(
                    new XElement("item"
                        , new XElement(CATEGORY_ELEMENT, (watchItem.Category == null || watchItem.Category == Category.Empty ? string.Empty : watchItem.Category.Id.ToString()))
                        , new XElement(DESCRIPTION_ELEMENT, watchItem.Description)
                        , new XElement(SEARCH_STRING_ELEMENT, watchItem.SearchString)
                        , new XElement(SEARCH_FOR_ALL_WORDS_ELEMENT, watchItem.SearchForAllWords)
                        , new XElement(SPOTTER_ELEMENT, watchItem.Spotter)
                        , new XElement(SUB_CATEGORY_ELEMENT, (watchItem.SubCategory == null || watchItem.SubCategory == SubCategory.Empty ? string.Empty : watchItem.SubCategory.Id.ToString()))
                        , new XElement(WHOLE_WORDS_ONLY_ELEMENT, watchItem.WholeWordsOnly)));
            }

            using (FileStream stream = new FileStream(_watchdogFile, FileMode.Create, FileAccess.ReadWrite))
            {
                repository.Save(stream);
            }
        }

        #endregion

        #region Private Methods

        private List<Category> GetCategories()
        {
            List<Category> categories = new List<Category>() { Category.Empty };

            //list with all categories
            categories.AddRange(from category in XDocument.Load(_categoriesFile).Root.Elements("category")
                                orderby category.Attribute("id").Value
                                select new Category()
                                {
                                    Name = category.Attribute("name").Value,
                                    Id = Int32.Parse(category.Attribute("id").Value),
                                    
                                    //list with all sub categories
                                    SubCategories = (from subCategory in category.Element("subcategories").Elements("subcategory")
                                                     orderby subCategory.Attribute("name").Value
                                                     select new SubCategory()
                                                     {
                                                         Id = Int32.Parse(subCategory.Attribute("id").Value),
                                                         Name = subCategory.Attribute("name").Value
                                                     }).ToList(),

                                    //list with all tags
                                    Tags = (from tag in category.Elements("tags").Elements("tag")
                                            select new Tag()
                                            {
                                                Id = tag.Attribute("id").Value,
                                                Name = tag.Attribute("name").Value,
                                            }).ToList()
                                });

            foreach (Category category in categories)
            {
                category.SubCategories.Insert(0, SubCategory.Empty);
            }

            return categories;
        }

        /// <summary>
        /// Get category based on id
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        private Category GetCategory(string id)
        {
            int categoryId;
            if (Int32.TryParse(id, out categoryId))
            {
                return Categories.SingleOrDefault((c) => c.Id == categoryId);
            }

            return Category.Empty;
        }

        /// <summary>
        /// Get sub category based on id
        /// </summary>
        /// <param name="element"></param>
        /// <returns></returns>
        private SubCategory GetSubCategory(Category category, XElement element)
        {
            if (element != null)
            {
                int subCategoryId;
                if (Int32.TryParse(element.Value, out subCategoryId))
                {
                    return category.SubCategories.SingleOrDefault((c) => c.Id == subCategoryId);
                }
            }

            return SubCategory.Empty;
        }

        private ObservableCollection<WatchItem> GetWatchItems()
        {
            if (File.Exists(_watchdogFile))
            {
                //load all watch items
                return new ObservableCollection<WatchItem>(
                        from item in XDocument.Load(_watchdogFile).Root.Elements("item")
                        let category = GetCategory(item.Element(CATEGORY_ELEMENT).Value)
                        orderby item.Element("description").Value
                        select new WatchItem()
                        {
                            Category = category,
                            Description = item.Element(DESCRIPTION_ELEMENT).Value,
                            SearchString = item.Element(SEARCH_STRING_ELEMENT).Value,
                            SearchForAllWords = Boolean.Parse(item.Element(SEARCH_FOR_ALL_WORDS_ELEMENT).Value),
                            Spotter = (item.Element(SPOTTER_ELEMENT) == null ? null : item.Element(SPOTTER_ELEMENT).Value),
                            SubCategory = GetSubCategory(category, item.Element(SUB_CATEGORY_ELEMENT)),
                            WholeWordsOnly = Boolean.Parse(item.Element(WHOLE_WORDS_ONLY_ELEMENT).Value),
                        });

            }

            return new ObservableCollection<WatchItem>();
        }

        #endregion

        #region Singleton
        
        private static volatile WatchdogRepository _instance;
        private static object _syncRoot = new object();

        public static WatchdogRepository Instance
        {
            get 
            {
                if (_instance == null) 
                {
                    lock (_syncRoot) 
                    {
                        if (_instance == null)
                        {
                            _instance = new WatchdogRepository();
                        }
                    }
                }

                return _instance;
            }
        }

        #endregion
    }
}
